from .solid_body import SolidBody


class BasicSolidBody(SolidBody):
    def __init__(
            self,
            id_color_dict,
            id_2d_list,
            width, height,
            col, row,
            vel_x, vel_y,
            acc_x, acc_y,
            temp, temp_freeze, temp_boil,
            density,
            color,
            name,
            flammability,
            state):
        super().__init__(
            id_color_dict,
            id_2d_list,
            width, height,
            col, row,
            vel_x, vel_y,
            acc_x, acc_y,
            temp, temp_freeze, temp_boil,
            density,
            color,
            name,
            flammability,
            state)

    def clone(self, col, row):
        return BasicSolidBody(
            self._id_color_dict,
            self._id_2d_list,
            self._width, self._height,
            col, row,
            self._vel_x, self._vel_y,
            self._acc_x, self._acc_y,
            self._temp, self._temp_freeze, self._temp_boil,
            self._density,
            self._color,
            self._name,
            self._flammability,
            self._state)

    def update_on_tick(self, driver, grid):
        self._update_position(grid)

    def _update_position(self, grid):
        can_move = True

        for row in range(self._height):
            for col in range(self._width):
                if self._particles[row][col] is None:
                    continue
                particle = self._particles[row][col]
                pos_probe = (particle.col + self._true_vel_x(), particle.row + self._true_vel_y())
                particle_probe = grid.get(pos_probe)
                if particle_probe is not None and self.contains(particle_probe):
                    continue
                else:
                    if not grid.is_in_bounds(pos_probe) or grid.exists(pos_probe):
                        can_move = False
                        break
            if not can_move:
                break

        if can_move:
            self._col += self._true_vel_x()
            self._row += self._true_vel_y()

            probe_dir_x = 1
            probe_start_col = 0
            probe_end_col = self._width

            probe_dir_y = -1
            probe_start_row = self._height - 1
            probe_end_row = -1

            if self._vel_x > 0.0:
                probe_dir_x = -1
                probe_start_col = self._width - 1
                probe_end_col = -1

            if self._vel_y < 0.0:
                probe_dir_y = 1
                probe_start_row = 0
                probe_end_row = self._height

            for row in range(probe_start_row, probe_end_row, probe_dir_y):
                for col in range(probe_start_col, probe_end_col, probe_dir_x):
                    if self._particles[row][col] is None:
                        continue
                    particle = self._particles[row][col]
                    pos = (particle.col, particle.row)
                    pos_new = (particle.col + self._true_vel_x(), particle.row + self._true_vel_y())
                    grid.swap(pos, pos_new)

        return can_move



from .particle import Particle


class SolidBodyParticle(Particle):
    def __init__(
            self,
            col, row,
            vel_x, vel_y,
            acc_x, acc_y,
            temp, temp_freeze, temp_boil,
            density,
            color,
            name,
            flammability,
            state):

        super().__init__(
            col, row,
            vel_x, vel_y,
            acc_x, acc_y,
            temp, temp_freeze, temp_boil,
            density,
            color,
            name,
            flammability,
            state)

    def clone(self, col, row):
        return SolidBodyParticle(
            col, row,
            self._vel_x, self._vel_y,
            self._acc_x, self._acc_y,
            self._temp, self._temp_freeze, self._temp_boil,
            self._density,
            self._color,
            self._name,
            self._flammability,
            self._state)

    def update_on_tick(self, driver, grid):
        pass

    def set_col(self, col):
        self._col = col

    def set_row(self, row):
        self._row = row



from . import screen
from .start_menu import *
from .driver import Driver
from .config import TOMENU_EVENT_TYPE


def main():
    pygame.init()
    clock = pygame.time.Clock()

    # Display the start menu and store play mode
    mode = display_start_menu()

    # Create and define screen regions
    (display, sandbox, element_menu, tool_menu) = screen.init_screen(mode)

    driver = Driver(mode, element_menu, display)

    if mode == "DISCOVERY":
        undiscovered = ["basalt", "steam", "fire", "water"]
        element_menu.discovery_init(undiscovered)
        driver.undiscovered.extend(undiscovered)

    while 1:
        # Blackout entire sandbox (should optimize in the future)
        pygame.draw.rect(display, SANDBOX_COLOR, sandbox)

        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                pygame.quit()
                sys.exit()
            elif event.type == TOMENU_EVENT_TYPE:
                return
            elif event.type == pygame.MOUSEBUTTONDOWN:
                if sandbox.collidepoint(pygame.mouse.get_pos()):
                    driver.set_tool_use(True)
                    if driver.get_tool() == "LINE" or driver.get_tool() == "RECT" or driver.get_tool() == "OVAL":
                        driver.start_shape(pygame.mouse.get_pos())
                elif element_menu.contains(event.pos[0], event.pos[1]):
                    element_menu.update(driver, event.pos[0], event.pos[1])
                elif tool_menu.contains(event.pos[0], event.pos[1]):
                    tool_menu.update(driver, event.pos[0], event.pos[1])
            elif event.type == pygame.MOUSEBUTTONUP:
                driver.set_tool_use(False)
                driver.end_active_shape()
                if sandbox.collidepoint(pygame.mouse.get_pos()):
                    if driver.get_tool() == "LINE":
                        driver.end_line(pygame.mouse.get_pos())
                    elif driver.get_tool() == "RECT":
                        driver.end_rect(pygame.mouse.get_pos())
                    elif driver.get_tool() == "OVAL":
                        driver.end_oval(pygame.mouse.get_pos())

        # Update particle positions and apply tool (if is being used)
        driver.update_particles(pygame.mouse)

        # Draw all particles in the sandbox
        driver.render(display)

        # Replace mouse pointer inside sandbox, otherwise show
        if sandbox.collidepoint(pygame.mouse.get_pos()):
            pygame.mouse.set_visible(False)
            driver.draw_tool_outline(pygame.mouse.get_pos(), sandbox, display)
        else:
            element_menu.draw_tooltip(pygame.mouse.get_pos()[0], pygame.mouse.get_pos()[1])
            pygame.mouse.set_visible(True)

        # Update and show FPS (used for debugging)
        screen.update_fps(display, clock)
        pygame.display.flip()
        clock.tick(FPS)


def loop_main():
    while 1:
        main()


if __name__ == '__main__':
    loop_main()



import random
import pygame
from .particle import Particle
from . import particle_data


class Gas(Particle):
    def __init__(
            self,
            col, row,
            vel_x, vel_y,
            acc_x, acc_y,
            temp, temp_freeze, temp_boil,
            density,
            color,
            name,
            flammability,
            state):
        super().__init__(
            col, row,
            vel_x, vel_y,
            acc_x, acc_y,
            temp, temp_freeze, temp_boil,
            density,
            color,
            name,
            flammability,
            state)

        self._lifespan = pygame.time.get_ticks() + 3000
        self._remaining = 3000

    def clone(self, col, row):
        return Gas(
            col, row,
            self._vel_x, self._vel_y,
            self._acc_x, self._acc_y,
            self._temp, self._temp_freeze, self._temp_boil,
            self._density,
            self._color,
            self._name,
            self._flammability,
            self._state)

    def update_on_tick(self, driver, grid):
        if (pygame.time.get_ticks() > self._lifespan) or (self._temp <= self._temp_freeze):
            driver.delete(self)
            return

        if self._needs_update is False:
            return

        min_vel_x = -1 if grid.is_in_bounds((self._col - 1, self._row)) else 0
        max_vel_x = 1 if grid.is_in_bounds((self._col + 1, self._row)) else 0

        if min_vel_x == max_vel_x:
            self._vel_x = min_vel_x
        else:
            self._vel_x = random.randrange(min_vel_x, max_vel_x + 1)

        self._update_vel()

        pos_path = self._get_positions_in_path(grid)

        if len(pos_path) == 0:
            self._needs_update = False

        for next_pos in pos_path:
            pos = (self._col, self._row)
            if grid.exists(next_pos) is False:
                self._force_update_near(grid)
                grid.swap(pos, next_pos)
            else:
                collider = grid.get(next_pos)

                if self.name != "void" and collider.name == "void":
                    driver.delete(self)
                    break

                # Heat Transfer
                near_list = grid.get_near((self._col, self._row))
                for particle in near_list:
                    temp_diff = (self._temp - particle.temp) / 50
                    particle.update_temp(particle.temp + temp_diff)
                    self.update_temp(self._temp - temp_diff)

                    # Metal -> lava when heated by fire
                    if particle.name == "metal" and particle.temp_freeze <= particle.temp:
                        oldtemp = particle.temp
                        particle.melt(driver, grid, particle_data.template_lava.clone(particle.col, particle.row))
                        particle.update_temp(oldtemp)

                    # Burning
                    if (particle.name == "powder" or particle.name == "wood") and (
                            particle.temp_freeze <= particle.temp):
                        oldtemp = particle.temp
                        particle.melt(driver, grid, particle_data.template_fire.clone(particle.col, particle.row))
                        particle.update_temp(oldtemp)

                if self._density > collider.density:
                    self._force_update_near(grid)
                    grid.swap(pos, next_pos)

                else:
                    pos_left = (self._col - 1, self._row)
                    pos_right = (self._col + 1, self._row)

                    can_go_left = grid.is_in_bounds(pos_left) \
                                  and grid.exists(pos_left) is False \
                                  or grid.is_in_bounds(pos_left) \
                                  and grid.exists(pos_left) \
                                  and self._density > grid.get(pos_left).density

                    can_go_right = grid.is_in_bounds(pos_right) \
                                   and grid.exists(pos_right) is False \
                                   or grid.is_in_bounds(pos_right) \
                                   and grid.exists(pos_right) \
                                   and self._density > grid.get(pos_right).density

                    min_vel_x = -1 if can_go_left else 0
                    max_vel_x = 1 if can_go_right else 0

                    if min_vel_x == max_vel_x == 0:
                        self._needs_update = False
                    else:
                        vel_x = random.randrange(min_vel_x, max_vel_x + 1)
                        next_pos = (self._col + vel_x, self._row)

                        self._force_update_near(grid)
                        grid.swap(pos, next_pos)
                break

    def save_lifespan(self):
        self._remaining = self._lifespan - pygame.time.get_ticks()

    def load_lifespan(self):
        self._lifespan = pygame.time.get_ticks() + self._remaining



from .config import *
from .particle import Particle

"""
    px_to_cell converts pixel coordinates into column or row coordinates, depending on whether you
    pass an x or a y value.
    This is useful in cases such as converting the mouse's x and y positions to their respective column and row
    counterparts in the grid.
"""


def px_to_cell(px):
    return int(px / PARTICLE_SIZE)


class Grid:
    def __init__(self):
        self.__num_cols = int(WINDOW_WIDTH / PARTICLE_SIZE)
        self.__num_rows = int(WINDOW_HEIGHT / PARTICLE_SIZE)
        self.__left = int(MARGIN / PARTICLE_SIZE)
        self.__right = int((MARGIN + SANDBOX_WIDTH) / PARTICLE_SIZE)
        self.__top = int((MARGIN * 2) / PARTICLE_SIZE)
        self.__bottom = int((MARGIN * 2 + SANDBOX_HEIGHT) / PARTICLE_SIZE)
        self.__cells = [[None for _ in range(self.__num_cols)] for _ in range(self.__num_rows)]

    """
        emplace adds a particle to its column and row, replacing the particle currently in that position.

        WARNING: This should not be used to add particles into the grid. Rather, the Driver's add method should be used,
        which calls emplace. If not, particles will not be added, updated, and rendered properly.
    """

    def emplace(self, particle):
        self.__cells[particle.row][particle.col] = particle

    def emplace_list(self, particle_list):
        for particle in particle_list:
            self.emplace(particle)

    def remove(self, particle):
        self.__cells[particle.row][particle.col] = None

    """
        swap switches the positions of two particles in the grid, also updating the positions of each particle.

        parameter lhs_pos is a tuple representing a column and row like so: (column, row)
        parameter rhs_pos is a tuple representing a column and row like so: (column, row)
    """

    def swap(self, lhs_pos, rhs_pos):
        lhs = self.__cells[lhs_pos[1]][lhs_pos[0]]
        rhs = self.__cells[rhs_pos[1]][rhs_pos[0]]

        self.__cells[lhs_pos[1]][lhs_pos[0]], self.__cells[rhs_pos[1]][rhs_pos[0]] = \
            self.__cells[rhs_pos[1]][rhs_pos[0]], self.__cells[lhs_pos[1]][lhs_pos[0]]

        if isinstance(lhs, Particle):
            lhs.set_pos(rhs_pos[0], rhs_pos[1])

        if isinstance(rhs, Particle):
            rhs.set_pos(lhs_pos[0], lhs_pos[1])

    """
        is_in_bounds returns whether a specified position is in bounds of the grid.

        parameter pos is a tuple representing a column and row like so: (column, row)
    """

    def is_in_bounds(self, pos):
        return self.__left <= pos[0] < self.__right and self.__top <= pos[1] < self.__bottom

    """
        exists returns whether there is a particle at the specified position.

        parameter pos is a tuple representing a column and row like so: (column, row)
    """

    def exists(self, pos):
        return self.__cells[pos[1]][pos[0]] is not None

    """
        is_in_bounds_and_exists combines the is_in_bounds and exists methods,
        returning whether the position is in bounds and there is a particle currently at the position.

        parameter pos is a tuple representing a column and row like so: (column, row)
    """

    def is_in_bounds_and_exists(self, pos):
        return self.is_in_bounds(pos) and self.exists(pos)

    """
        get returns the particle at the specified position, or None if there is no particle at the specified position.
        It is useful to check the is_in_bounds and exists methods to determine whether there is a particle at the
        position prior to calling the get method.

        parameter pos is a tuple representing a column and row like so: (column, row)
    """

    def get(self, pos):
        return self.__cells[pos[1]][pos[0]]

    """
        get_near returns the four particles around the particle at a specified position.
        the particles returned are to the:
            top,
            bottom,
            left,
            right
        of the specified position.

        parameter pos is a tuple representing a column and row like so: (column, row)
    """

    def get_near(self, pos):
        near_list = []
        col = pos[0]
        row = pos[1]

        self.__push_if_able(near_list, (col, row - 1))
        self.__push_if_able(near_list, (col, row + 1))
        self.__push_if_able(near_list, (col - 1, row))
        self.__push_if_able(near_list, (col + 1, row))

        return near_list

    """
        __push_if_able is a helper method for the get_near method that pushes the particle at the designated position
        to the specified list, if the position is in bounds and there is a particle there.

        parameter target_list is a list to which the particle currently at the designated position will be pushed,
            if the position is in bounds and there is currently a particle there.
        parameter pos is a tuple representing a column and row like so: (column, row)
    """

    def __push_if_able(self, target_list, pos):
        if self.is_in_bounds_and_exists(pos):
            target_list.append(self.get(pos))

    @property
    def left(self):
        return self.__left

    @property
    def right(self):
        return self.__right

    @property
    def top(self):
        return self.__top

    @property
    def bottom(self):
        return self.__bottom



from . import solid
from . import liquid
from . import gas
from . import fixed
from . import basic_solid_body
from . import ufo_solid_body


"""
    Template particles that can be used for the Painter, or elsewhere.
"""

template_sand = solid.Solid(
                    0, 0,                   # position
                    0.0, 1.0,               # velocity
                    0.0, 0.0,               # acceleration
                    20, 1000, 2230,         # temperature (temp, freeze, boil)
                    1.6,                    # density
                    (255, 255, 0),          # color
                    "sand",                 # name
                    0,                      # flammability
                    "solid")                # state (solid, liquid, gas, fixed)

template_water = liquid.Liquid(
                    0, 0,                   # position
                    0.0, 1.0,               # velocity
                    0.0, 0.0,               # acceleration
                    20, 0, 100,             # temperature
                    1.0,                    # density
                    (0, 0, 255),            # color
                    "water",                # name
                    0,                      # flammability
                    "liquid")               # state

template_lava = liquid.Liquid(
                    0, 0,                   # position
                    0.0, 1.0,               # velocity
                    0.0, 0.0,               # acceleration
                    2000, 800, 10000,       # temperature
                    3.0,                    # density
                    (153, 0, 0),            # color
                    "lava",                 # name
                    0,                      # flammability
                    "liquid")               # state

template_steam = gas.Gas(
                    0, 0,                   # position
                    0.0, -1.0,              # velocity
                    0.0, 0.0,               # acceleration
                    10, -100, 10000,        # temperature
                    0.5,                    # density
                    (125, 125, 125),        # color
                    "steam",                # name
                    0.95,                   # flammability
                    "gas")                  # state

template_wood = fixed.Fixed(
                    0, 0,                   # position
                    0.0, 0.0,               # velocity
                    0.0, 0.0,               # acceleration
                    20, 400, 7000,          # temperature
                    5.0,                    # density
                    (160, 82, 45),          # color
                    "wood",                 # name
                    0.2,                    # flammability
                    "fixed")                # state

template_metal = fixed.Fixed(
                    0, 0,                   # position
                    0.0, 0.0,               # velocity
                    0.0, 0.0,               # acceleration
                    20, 1200, 10000,        # temperature
                    15.0,                   # density
                    (192, 192, 192),        # color
                    "metal",                # name
                    0.1,                    # flammability
                    "fixed")                # state

template_basalt = fixed.Fixed(
                    0, 0,                   # position
                    0.0, 0.0,               # velocity
                    0.0, 0.0,               # acceleration
                    600, 1500, 800,         # temperature
                    15.0,                   # density
                    (53, 72, 96),           # color
                    "basalt",               # name
                    0.1,                    # flammability
                    "fixed")                # state

template_fire = gas.Gas(
                    0, 0,                   # position
                    0.0, -1.0,              # velocity
                    0.0, 0.0,               # acceleration
                    800, -300, 10000,       # temperature
                    0.5,                    # density
                    (201, 75, 44),          # color
                    "fire",                 # name
                    0,                      # flammability
                    "gas")                  # state

template_powder = solid.Solid(
                    0, 0,                   # position
                    0.0, 1.0,               # velocity
                    0.0, 0.0,               # acceleration
                    20, 300, 2230,          # temperature (temp, freeze, boil)
                    1,                      # density
                    (249, 218, 159),        # color
                    "powder",               # name
                    0.9,                    # flammability
                    "solid")                # state

template_snow = solid.Solid(
                    0, 0,                   # position
                    0.0, 1.0,               # velocity
                    0.0, 0.0,               # acceleration
                    -30, 0, 100,            # temperature (temp, freeze, boil)
                    .9,                     # density
                    (255, 250, 250),        # color
                    "snow",                 # name
                    0,                      # flammability
                    "solid")                # state

template_ice = fixed.Fixed(
                    0, 0,                   # position
                    0.0, 0.0,               # velocity
                    0.0, 0.0,               # acceleration
                    -80, 0, 100,            # temperature
                    1,                      # density
                    (47, 230, 239),         # color
                    "ice",                  # name
                    0,                      # flammability
                    "fixed")                # state

template_oil = liquid.Liquid(
                    0, 0,                   # position
                    0.0, 1.0,               # velocity
                    0.0, 0.0,               # acceleration
                    20, -100, 100,          # temperature
                    0.3,                    # density
                    (86, 63, 43),           # color
                    "oil",                  # name
                    0.98,                   # flammability
                    "liquid")               # state

template_fog = gas.Gas(
                    0, 0,                   # position
                    0.0, 0.0,               # velocity
                    0.0, 0.0,               # acceleration
                    20, -100, 100,          # temperature
                    0.3,                    # density
                    (218, 234, 232),        # color
                    "fog",                  # name
                    0,                      # flammability
                    "gas")                  # state

template_stone = solid.Solid(
                    0, 0,                   # position
                    0.0, 1.0,               # velocity
                    0.0, 0.0,               # acceleration
                    20, 1100, 2230,         # temperature (temp, freeze, boil)
                    20,                     # density
                    (137, 137, 137),        # color
                    "stone",                # name
                    0,                      # flammability
                    "solid")                # state
                    
template_generator = fixed.Fixed(
                    0, 0,                   # position
                    0.0, 0.0,               # velocity
                    0.0, 0.0,               # acceleration
                    20, -100, 1000,         # temperature
                    15.0,                   # density
                    (0, 153, 153),          # color
                    "Water Generator",      # name
                    0,                      # flammability
                    "fixed")                # state


template_void = fixed.Fixed(
                    0, 0,                   # position
                    0.0, 0.0,               # velocity
                    0.0, 0.0,               # acceleration
                    20, -100, 1000,         # temperature
                    15.0,                   # density
                    (3, 0, 71),             # color
                    "void",                 # name
                    0,                      # flammability
                    "fixed")                # state

template_acid = liquid.Liquid(
                    0, 0,                   # position
                    0.0, 1.0,               # velocity
                    0.0, 0.0,               # acceleration
                    20, -100, 100,          # temperature
                    0.5,                    # density
                    (29, 224, 48),          # color
                    "acid",                 # name
                    0,                      # flammability
                    "liquid")               # state

template_bush = fixed.Fixed(
                    0, 0,                   # position
                    0.0, 0.0,               # velocity
                    0.0, 0.0,               # acceleration
                    20, -100, 400,          # temperature
                    0.5,                    # density
                    (17, 96, 22),           # color
                    "bush",                 # name
                    0.75,                   # flammability
                    "fixed")                # state

template_vine = fixed.Fixed(
                    0, 0,                   # position
                    0.0, 0.0,               # velocity
                    0.0, 0.0,               # acceleration
                    20, -100, 400,          # temperature
                    0.5,                    # density
                    (68, 214, 78),          # color
                    "vine",                 # name
                    0.75,                   # flammability
                    "fixed")                # state

template_solidbody_test = basic_solid_body.BasicSolidBody(
    {
        0: None,
        1: (255, 0, 0),
        2: (0, 255, 0),
        3: (0, 0, 255),
        4: (255, 255, 0),
        5: (0, 255, 255),
        6: (255, 0, 255)
    },
    [
        [0, 0, 1, 0, 0],
        [0, 3, 2, 3, 0],
        [6, 5, 4, 5, 6],
        [0, 3, 2, 3, 0],
        [0, 0, 1, 0, 0]
    ],
    5, 5,
    0, 0,
    0.0, -1.0,
    0.0, 0.0,
    20, 1000, 2230,
    1000.0,
    (255, 0, 255),
    "test",
    0,
    "solid")

template_solidbody_smiley = basic_solid_body.BasicSolidBody(
    {
        0: None,
        1: (255, 255, 0)
    },
    [
        [0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0],
        [0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0],
        [0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0],
        [0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0],
        [0, 1, 1, 1, 1, 0, 1, 1, 1, 1, 0, 1, 1, 1, 1, 0],
        [1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1],
        [1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1],
        [1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1],
        [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
        [1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1],
        [1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1],
        [0, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 0],
        [0, 1, 1, 1, 0, 0, 1, 1, 1, 1, 0, 0, 1, 1, 1, 0],
        [0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0],
        [0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0],
        [0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0],
    ],
    16, 16,
    0, 0,
    0.0, 1.0,
    0.0, 0.0,
    20, 600, 2230,
    1000.0,
    (255, 255, 0),
    "smiley",
    0,
    "solid")

template_ufo = ufo_solid_body.UFOSolidBody(
    {
        0: None,
        1: (125, 125, 125),
        2: (225, 225, 225),
        3: (0, 125, 0),
        4: (255, 255, 0)
    },
    [
        [0, 0, 0, 2, 2, 2, 0, 0, 0],
        [0, 0, 2, 2, 3, 2, 2, 0, 0],
        [1, 1, 1, 1, 1, 1, 1, 1, 1],
        [0, 1, 4, 1, 1, 1, 4, 1, 0]
    ],
    9, 4,
    0, 0,
    1.0, 0.0,
    0.0, 0.0,
    20, 1000, 2230,
    1000.0,
    (0, 255, 0),
    "ufo",
    0,
    "solid")

# Particle dictionary used to create particle selection tool
ELEMENTS = {
    "FIXED": [template_wood, template_metal, template_basalt, template_ice, template_bush, template_vine],
    "SOLIDS": [template_sand, template_powder, template_stone, template_snow],
    "LIQUIDS": [template_water, template_lava, template_oil, template_acid],
    "GASES": [template_steam, template_fire, template_fog],
    "SPECIAL": [template_generator, template_void],
    "SOLID BODIES": [template_solidbody_test, template_solidbody_smiley, template_ufo]
}



import random
from .particle import Particle
from . import particle_data


class Liquid(Particle):
    def __init__(
            self,
            col, row,
            vel_x, vel_y,
            acc_x, acc_y,
            temp, temp_freeze, temp_boil,
            density,
            color,
            name,
            flammability,
            state):
        super().__init__(
            col, row,
            vel_x, vel_y,
            acc_x, acc_y,
            temp, temp_freeze, temp_boil,
            density,
            color,
            name,
            flammability,
            state)

    def clone(self, col, row):
        return Liquid(
            col, row,
            self._vel_x, self._vel_y,
            self._acc_x, self._acc_y,
            self._temp, self._temp_freeze, self._temp_boil,
            self._density,
            self._color,
            self._name,
            self._flammability,
            self._state)

    def update_on_tick(self, driver, grid):
        if self._needs_update is False:
            return

        pos_path = self._get_positions_in_path(grid)

        if len(pos_path) == 0:
            self._needs_update = False

        for next_pos in pos_path:
            pos = (self._col, self._row)
            if grid.exists(next_pos) is False:
                self._force_update_near(grid)
                grid.swap(pos, next_pos)
            else:
                collider = grid.get(next_pos)

                if self.name != "void" and collider.name == "void":
                    driver.delete(self)
                    break

                if self.name == "acid" and collider.name != "acid":
                    driver.delete(collider)
                    driver.delete(self)
                    break

                # Heat transfer
                near_list = grid.get_near((self._col, self._row))
                for particle in near_list:

                    temp_diff = (self._temp - particle.temp) / 50
                    particle.update_temp(particle.temp + temp_diff)
                    self.update_temp(self._temp - temp_diff)

                    # Water below freezing -> ice
                    if particle.name == "water" and particle.temp_freeze > particle.temp:
                        oldtemp = particle.temp
                        particle.freeze(driver, grid, particle_data.template_ice.clone(particle.col, particle.row))
                        particle.update_temp(oldtemp)

                    # Lava -> basalt when cooled
                    if particle.name == "lava" and particle.temp_freeze >= particle.temp:
                        oldtemp = particle.temp
                        particle.freeze(driver, grid, particle_data.template_basalt.clone(particle.col, particle.row))
                        particle.update_temp(oldtemp)

                    # Smiley particles -> lava when melted
                    if particle.name == "smiley" and particle.temp_freeze < particle.temp:
                        oldtemp = particle.temp
                        particle.melt(driver, grid, particle_data.template_lava.clone(particle.col, particle.row))
                        particle.update_temp(oldtemp)

                    # UFO particles -> lava when melted
                    if particle.name == "ufo" and particle.temp_freeze < particle.temp:
                        oldtemp = particle.temp
                        particle.melt(driver, grid, particle_data.template_lava.clone(particle.col, particle.row))
                        particle.update_temp(oldtemp)

                # Water -> ice when below freezing
                if self.name == "water" and self._temp_freeze > self._temp:
                    oldtemp = self._temp
                    self.freeze(driver, grid, particle_data.template_ice.clone(self._col, self._row))
                    self.update_temp(oldtemp)

                # All liquids except oil above boiling -> gas
                if self.name != "oil" and self._temp_boil <= self._temp:
                    oldtemp = self._temp
                    self.boil(driver, grid, particle_data.template_steam.clone(self._col, self._row))
                    self.update_temp(oldtemp)

                # Oil burns
                if self.name == "oil" and self._temp_boil <= self._temp:
                    oldtemp = self._temp
                    self.boil(driver, grid, particle_data.template_fire.clone(self._col, self._row))
                    self.update_temp(oldtemp)

                # Lava -> basalt when cooled
                if self.name == "lava" and self._temp_freeze >= self._temp:
                    oldtemp = self._temp
                    self.freeze(driver, grid, particle_data.template_basalt.clone(self._col, self._row))
                    self.update_temp(oldtemp)

                if self.density > collider.density:
                    self._force_update_near(grid)
                    grid.swap(pos, next_pos)
                else:
                    pos_left = (self._col - 1, self._row)
                    pos_right = (self._col + 1, self._row)

                    can_go_left = grid.is_in_bounds(pos_left)           \
                        and grid.exists(pos_left) is False              \
                        or grid.is_in_bounds(pos_left)                  \
                        and grid.exists(pos_left)                       \
                        and self._density > grid.get(pos_left).density

                    can_go_right = grid.is_in_bounds(pos_right)         \
                        and grid.exists(pos_right) is False             \
                        or grid.is_in_bounds(pos_right)                 \
                        and grid.exists(pos_right)                      \
                        and self._density > grid.get(pos_right).density

                    min_vel_x = -1 if can_go_left else 0
                    max_vel_x = 1 if can_go_right else 0

                    if min_vel_x == max_vel_x == 0:
                        self._needs_update = False
                    else:
                        vel_x = random.randrange(min_vel_x, max_vel_x + 1)
                        next_pos = (self._col + vel_x, self._row)

                        self._force_update_near(grid)
                        grid.swap(pos, next_pos)
                break



import sys
from .config import *


def create_menu_button(menu, text, y):
    button = pygame.Rect(0, 0, 400, 45)
    button.center = (MENU_WIDTH/2, y)
    pygame.draw.rect(menu, (180, 180, 180), button)
    pygame.draw.rect(menu, (255, 255, 255), button, 1)

    font = pygame.font.Font(FONT_PATH, 36)
    label = font.render(text, False, (0, 0, 0))
    menu.blit(label, (button.x + ((button.width-label.get_width()) / 2), button.y))

    return button


def display_start_menu():
    clock = pygame.time.Clock()
    menu = pygame.display.set_mode((MENU_WIDTH, MENU_HEIGHT))

    menu.fill(BG_COLOR)

    title_font = pygame.font.Font(FONT_PATH, 60)
    title = title_font.render("SANDCRAFT", False, (255, 255, 255))
    title_rect = title.get_rect()
    title_rect.center = (MENU_WIDTH/2, MENU_HEIGHT/3)
    menu.blit(title, title_rect)

    y = math.floor((1/2) * MENU_HEIGHT)
    increment = 60
    sandbox_button = create_menu_button(menu, "SANDBOX MODE", y)
    discovery_button = create_menu_button(menu, "DISCOVERY MODE", y + increment)
    load_button = create_menu_button(menu, "LOAD", y + 2 * increment)
    quit_button = create_menu_button(menu, "QUIT GAME", y + 3 * increment)

    while 1:
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                pygame.quit()
                sys.exit()
            if event.type == pygame.MOUSEBUTTONDOWN:
                if sandbox_button.collidepoint(pygame.mouse.get_pos()):
                    return "SANDBOX"
                elif discovery_button.collidepoint(pygame.mouse.get_pos()):
                    return "DISCOVERY"
                elif load_button.collidepoint(pygame.mouse.get_pos()):
                    return "LOAD"
                elif quit_button.collidepoint(pygame.mouse.get_pos()):
                    pygame.quit()
                    sys.exit()
        pygame.display.update()
        clock.tick(15)



import abc
import math


class GridObject(metaclass=abc.ABCMeta):
    def __init__(
            self,
            col, row,
            vel_x, vel_y,
            acc_x, acc_y,
            temp, temp_freeze, temp_boil,
            density,
            color,
            name,
            flammability,
            state):

        self._col = col
        self._row = row
        self._vel_x = vel_x
        self._vel_y = vel_y
        self._acc_x = acc_x
        self._acc_y = acc_y
        self._temp = temp
        self._temp_freeze = temp_freeze
        self._temp_boil = temp_boil
        self._density = density
        self._color = color
        self._name = name
        self._flammability = flammability
        self._state = state

    @abc.abstractmethod
    def clone(self, col, row):
        pass

    @abc.abstractmethod
    def emplace(self, driver):
        pass

    """
        _true_vel_x returns the value that the col will be updated by every tick (conversion from float to int)
    """
    def _true_vel_x(self):
        return int(math.floor(self._vel_x))

    """
        _true_vel_y returns the value that the row will be updated by every tick (conversion from float to int)
    """
    def _true_vel_y(self):
        return int(math.floor(self._vel_y))

    @property
    def col(self):
        return self._col

    @property
    def row(self):
        return self._row

    @property
    def temp(self):
        return self._temp

    @property
    def density(self):
        return self._density

    @property
    def color(self):
        return self._color

    @property
    def name(self):
        return self._name

    @property
    def state(self):
        return self._state

    @property
    def temp_freeze(self):
        return self._temp_freeze



import abc
import pygame
from .config import PARTICLE_SIZE
from .grid_object import GridObject


class Particle(GridObject, metaclass=abc.ABCMeta):
    def __init__(
            self,
            col, row,
            vel_x, vel_y,
            acc_x, acc_y,
            temp, temp_freeze, temp_boil,
            density,
            color,
            name,
            flammability,
            state):

        super().__init__(
            col, row,
            vel_x, vel_y,
            acc_x, acc_y,
            temp, temp_freeze, temp_boil,
            density,
            color,
            name,
            flammability,
            state)

        self._is_live = True
        self._needs_update = True

    """
        clone is an abstract method that is overridden by state classes deriving from Particle.
        This function returns a clone of the callee.
        This function is useful in cases such as the Painter class, which holds a template particle that
        is cloned to the mouse's current column and row in the grid when the left mouse button is pressed.
    """
    @abc.abstractmethod
    def clone(self, col, row):
        pass

    """
        update_on_tick is an abstract method that is overridden by state classes deriving from Particle.
        This is the function that handles the particle's movement and interactions.
    """
    @abc.abstractmethod
    def update_on_tick(self, driver, grid):
        pass

    def emplace(self, driver):
        driver.add(self)

    """
        render draws the particle to the screen using a primitive pygame Rect
    """
    def render(self, screen):
        pygame.draw.rect(
            screen,
            self._color,
            pygame.Rect(
                self._col * PARTICLE_SIZE,
                self._row * PARTICLE_SIZE,
                PARTICLE_SIZE,
                PARTICLE_SIZE))

    """
        force_update will flip the particle's _needs_update boolean flag to True, which will result in the
        particle being updated the next tick.
        This is necessary, because particles will get stuck and/or not interact properly if their update
        flag is False, as it will never be reset.
        This method is generally called by other particles during their own update_on_tick cycles.
    """
    def force_update(self):
        self._needs_update = True

    def deny_update(self):
        self._needs_update = False

    """
        set_pos changes the particle's current col and row in the grid
    """
    def set_pos(self, col, row):
        self._col = col
        self._row = row

    """
        __boil is defined to reduce redundancy between the state classes (solid, liquid, gas, etc.)
        
        Freezing and melting are identical in function for now.
    """
    def boil(self, driver, grid, new_particle):
        self._is_live = False
        near_list = grid.get_near((self._col, self._row))
        for particle in near_list:
            particle.force_update()
        driver.add(new_particle)

    def freeze(self, driver, grid, new_particle):
        self._is_live = False
        near_list = grid.get_near((self._col, self._row))
        for particle in near_list:
            particle.force_update()
        driver.add(new_particle)

    def melt(self, driver, grid, new_particle):
        self._is_live = False
        near_list = grid.get_near((self._col, self._row))
        for particle in near_list:
            particle.force_update()
        driver.add(new_particle)

    """
        __force_update_near is defined to reduce redundancy between the state classes (solid, liquid, gas, etc.)
    """
    def _force_update_near(self, grid):
        near_list = grid.get_near((self._col, self._row))
        for particle in near_list:
            particle.force_update()

    """
        _update_vel updates the current velocity according to the acceleration.
    """
    def _update_vel(self):
        self._vel_x += self._acc_x
        self._vel_y += self._acc_y

    """
        _get_particles_in_path returns a list of particles that have possible collisions upon the next position update.
    """
    def _get_particles_in_path(self, grid):
        in_path = []

        positions_in_path_list = self._get_positions_in_path(grid)

        for pos in positions_in_path_list:
            if grid.exists(pos):
                in_path.append(grid.get(pos))

        return in_path

    """
        _get_positions_in_path returns a list of positions that have possible collisions upon the next position update.
    """
    def _get_positions_in_path(self, grid):
        in_path = []

        probe_x = self._col
        probe_y = self._row
        final_x = self._col + self._true_vel_x()
        final_y = self._row + self._true_vel_y()

        if final_x < grid.left:
            final_x = grid.left
        elif final_x >= grid.right:
            final_x = grid.right - 1

        if final_y < grid.top:
            final_y = grid.top
        elif final_y >= grid.bottom:
            final_y = grid.bottom - 1

        step_x = -1 if probe_x > final_x else 1
        step_y = -1 if probe_y > final_y else 1

        while probe_x != final_x or probe_y != final_y:
            if probe_x != final_x:
                probe_x += step_x

            if probe_y != final_y:
                probe_y += step_y

            in_path.append((probe_x, probe_y))

        return in_path

    """
        _update_temp changes the temperature of the particle, for collision purposes
    """
    def update_temp(self, new_temp):
        self._temp = new_temp

    """
        properties (read only)
    """

    @property
    def is_live(self):
        return self._is_live

    @property
    def needs_update(self):
        return self._needs_update

    def get_temp(self):
        return self._temp

    def set_temp(self, temp):
        self._temp = temp

    temp = property(get_temp, set_temp)



from .config import *


class ToolMenu:
    def __init__(self, surface, x, y, width):
        self._surface = surface
        self._x = x
        self._y = y
        self._width = width
        self._height = 100
        self._TOOLS = ["ADD", "DELETE", "ERASE", "LINE", "RECT",
                       "OVAL", "INSPECT", "CLEAR", "REMOVE", "SAVE",
                       "LOAD", "EXIT"]
        self.tool_buttons = []
        self._rows = 0
        self.draw()

    def draw(self):
        button_x = self._x
        button_y = self._y
        rows = 1
        for tool in self._TOOLS:
            new_button = self.ToolButton(self._surface, button_x, button_y, tool)

            # If new button is outside of tool menu
            if button_x + new_button.get_width() > self._width:  # + self._x:
                pygame.draw.rect(self._surface, BG_COLOR,
                                 pygame.Rect(button_x, button_y, new_button.get_width(), new_button.get_height()))
                button_x = self._x
                button_y = new_button.get_height() + 10 + button_y  # self._y +
                new_button = self.ToolButton(self._surface, button_x, button_y, tool)
                rows += 1

            button_x += new_button.get_width() + 10
            self.tool_buttons.append(new_button)

        self.draw_adjustment(1, rows)
        self._rows = rows

    # Checks if a button was clicked, then changes corresponding button to active
    def update(self, driver, x, y):
        for b in self.tool_buttons:
            if b.contains(x, y):
                for button in self.tool_buttons:
                    button.set_inactive()
                    if button.contains(x, y):
                        if button.get_tool() in ["ADD", "DELETE", "ERASE", "LINE", "RECT", "OVAL", "INSPECT"]:
                            button.set_active(driver)
                        elif button.get_tool() == "CLEAR":
                            driver.clear_sandbox()
                        elif button.get_tool() == "REMOVE":
                            driver.clear_element()
                        elif button.get_tool() == "SAVE":
                            driver.save_state()
                        elif button.get_tool() == "LOAD":
                            driver.load_state()
                        elif button.get_tool() == "EXIT":
                            pygame.event.post(TOMENU_EVENT)
                        elif button.get_tool() == "-":
                            driver.set_size(-1)
                            self.update_tool_size(driver)
                        elif button.get_tool() == "+":
                            driver.set_size(1)
                            self.update_tool_size(driver)
                    button.update()

    def contains(self, x, y):
        if x < self._x or self._x + self._width < x:
            return False
        if y < self._y or self._y + self._height < y:
            return False
        return True

    def update_tool_size(self, driver):
        tool_size = driver.get_size()
        font = pygame.font.Font(FONT_PATH, 11)
        label = font.render(f"BRUSH SIZE: {tool_size} ", True, pygame.Color(255, 255, 255), BG_COLOR)
        top = self._y + 35 * self._rows
        left = self._x
        self._surface.blit(label, (left, top))

    def draw_adjustment(self, tool_size, rows):
        top = self._y + 35 * rows
        left = self._x

        font = pygame.font.Font(FONT_PATH, 11)
        label = font.render(f"BRUSH SIZE: {tool_size}", True, pygame.Color(255, 255, 255), BG_COLOR)
        self._surface.blit(label, (left, top))

        top += 20

        button = self.ToolButton(self._surface, left, top, "-")
        left += button.get_width() + 10
        self.tool_buttons.append(button)

        button = self.ToolButton(self._surface, left, top, "+")
        self.tool_buttons.append(button)

        self._height += 100

    class ToolButton:
        def __init__(self, surface, x, y, name):
            self._surface = surface
            self._x = x
            self._y = y
            self._width = 0
            self._height = 0
            self._name = name
            self._active = False
            self.create()
            self.update()

        # Redraws button, returns bounding Rect for refresh
        def create(self):
            font = pygame.font.Font(FONT_PATH, 11)
            label = font.render(self._name, True, pygame.Color(0, 0, 0))
            button = label.get_rect()
            button.update(self._x, self._y, button.width + 20, button.height + 10)

            self._width = button.width
            self._height = button.height

            pygame.draw.rect(self._surface, (180, 180, 180), button)
            self._surface.blit(label, (self._x + 10, self._y + 5))

        def update(self):
            if self._active:
                pygame.draw.lines(
                    self._surface, (255, 0, 0), True, ((self._x, self._y),
                                                       (self._x + self._width-1, self._y),
                                                       (self._x + self._width-1, self._y + self._height-1),
                                                       (self._x, self._y + self._height-1)))
            else:
                pygame.draw.lines(
                    self._surface, (0, 0, 0), True, ((self._x, self._y),
                                                     (self._x + self._width - 1, self._y),
                                                     (self._x + self._width - 1, self._y + self._height - 1),
                                                     (self._x, self._y + self._height - 1)))

        def get_width(self):
            return self._width

        def get_height(self):
            return self._height

        def get_tool(self):
            return self._name

        def set_active(self, driver):
            self._active = True
            driver.set_tool(self._name)

        def set_inactive(self):
            self._active = False

        def contains(self, x, y):
            if x < self._x or self._x + self._width < x:
                return False

            if y < self._y or self._y + self._height < y:
                return False

            return True



from .particle import Particle
from . import particle_data


class Solid(Particle):
    def __init__(
            self,
            col, row,
            vel_x, vel_y,
            acc_x, acc_y,
            temp, temp_freeze, temp_boil,
            density,
            color,
            name,
            flammability,
            state):
        super().__init__(
            col, row,
            vel_x, vel_y,
            acc_x, acc_y,
            temp, temp_freeze, temp_boil,
            density,
            color,
            name,
            flammability,
            state)

    def clone(self, col, row):
        return Solid(
            col, row,
            self._vel_x, self._vel_y,
            self._acc_x, self._acc_y,
            self._temp, self._temp_freeze, self._temp_boil,
            self._density,
            self._color,
            self._name,
            self._flammability,
            self._state)

    def update_on_tick(self, driver, grid):
        if self._needs_update is False:
            return

        self._update_vel()

        pos_path = self._get_positions_in_path(grid)

        if len(pos_path) == 0:
            self._needs_update = False

        for next_pos in pos_path:
            pos = (self._col, self._row)

            if grid.exists(next_pos) is False:
                self._force_update_near(grid)
                grid.swap(pos, next_pos)
            else:
                collider = grid.get(next_pos)

                if self.name != "void" and collider.name == "void":
                    driver.delete(self)
                    break

                # Heat transfer
                near_list = grid.get_near((self._col, self._row))
                for particle in near_list:

                    temp_diff = (self._temp - particle.temp) / 50
                    if particle.name == "fire":
                        temp_diff = temp_diff * self._flammability
                    particle.update_temp(particle.temp + temp_diff)
                    self.update_temp(self._temp - temp_diff)

                # Burning
                if (self.name == "powder" or self.name == "wood") and (self._temp_freeze <= self._temp):
                    oldtemp = self._temp
                    self.melt(driver, grid, particle_data.template_fire.clone(self._col, self._row))
                    self.update_temp(oldtemp)

                # Molten stone or sand -> lava
                if (self.name == "stone" or self.name == "sand") and self._temp_freeze <= self._temp:
                    oldtemp = self._temp
                    self.melt(driver, grid, particle_data.template_lava.clone(self._col, self._row))
                    self.update_temp(oldtemp)

                # snow melts into water
                if self.name == "snow" and self._temp_freeze <= self._temp:
                    oldtemp = self._temp
                    self.melt(driver, grid, particle_data.template_water.clone(self._col, self._row))
                    self.update_temp(oldtemp)

                if self._density > collider.density:
                    self._force_update_near(grid)
                    grid.swap(pos, next_pos)
                else:
                    left_pos = (self._col - 1, self._row + 1)
                    right_pos = (self._col + 1, self._row + 1)

                    left_in_bounds = grid.is_in_bounds(left_pos)
                    right_in_bounds = grid.is_in_bounds(right_pos)

                    left_exists = left_in_bounds and grid.exists(left_pos)
                    right_exists = right_in_bounds and grid.exists(right_pos)

                    if left_exists is False and left_in_bounds \
                            or left_exists is True and self._density > grid.get(left_pos).density:
                        self._force_update_near(grid)

                        grid.swap(pos, left_pos)

                    elif right_exists is False and right_in_bounds \
                            or right_exists is True and self._density > grid.get(right_pos).density:
                        self._force_update_near(grid)

                        grid.swap(pos, right_pos)

                    else:
                        self._needs_update = False
                break



from .grid import px_to_cell


class Painter:
    def __init__(self, template_particle):
        self.__temp = template_particle

    def use_tool(self, pygame_mouse, driver, grid):
        mouse_col = px_to_cell(pygame_mouse.get_pos()[0])
        mouse_row = px_to_cell(pygame_mouse.get_pos()[1])

        # Add particles to grid based on tool size
        for x in range(mouse_col, mouse_col + driver.get_size()):
            for y in range(mouse_row, mouse_row + driver.get_size()):
                if grid.is_in_bounds([x, y]):
                    if grid.exists([x, y]) is False and driver.get_tool() == "ADD":
                        self.__temp.clone(x, y).emplace(driver)
                    elif grid.exists([x, y]) is True and driver.get_tool() == "DELETE":
                        driver.delete(grid.get([x, y]))
                    elif grid.exists([x, y]) is True and driver.get_tool() == "ERASE":
                        if grid.get([x, y]).name == self.__temp.name:
                            driver.delete(grid.get([x, y]))

    def set_template_particle(self, template_particle):
        self.__temp = template_particle

    def get_template_particle(self):
        return self.__temp



from .element_menu import *
from .tool_menu import *


def init_screen(mode):
    pygame.display.set_caption('Sandcraft')
    surface = pygame.display.set_mode((WINDOW_WIDTH, WINDOW_HEIGHT))

    tbar_height = math.floor(SANDBOX_HEIGHT * 0.3)
    tbar_width = (SANDBOX_WIDTH / 2) - (MARGIN / 2)

    pbar_height, pbar_width = tbar_height, tbar_width

    pbar_top = 3 * MARGIN + SANDBOX_HEIGHT
    tbar_top = pbar_top
    tbar_left = MARGIN
    pbar_left = 2 * MARGIN + tbar_width

    surface.fill(BG_COLOR)

    sandbox = pygame.Rect(MARGIN, 2 * MARGIN, SANDBOX_WIDTH, SANDBOX_HEIGHT)
    surface.fill(SANDBOX_COLOR, sandbox)

    # Top Menu Bar
    title_font = pygame.font.Font(FONT_PATH, 22)
    title_text = title_font.render("SANDCRAFT", True, pygame.Color(255, 255, 255))
    surface.blit(title_text, (MARGIN, (2*MARGIN - 22)/3))

    # Particles Selection
    tool_menu = ToolMenu(surface, tbar_left, tbar_top, tbar_width)
    element_menu = ElementMenu(surface, pbar_left, pbar_top, pbar_width, mode)

    return [surface, sandbox, element_menu, tool_menu]


def in_sandbox(x, y):
    if x < MARGIN or MARGIN + SANDBOX_WIDTH < x:
        return False
    if y < MARGIN * 2 or MARGIN * 2 + SANDBOX_HEIGHT < y:
        return False
    return True


def update_fps(display, clock):
    fps_bg = pygame.Rect(WINDOW_WIDTH - 50, WINDOW_HEIGHT - 15, 50, 15)
    pygame.draw.rect(display, BG_COLOR, fps_bg)

    fps = int(clock.get_fps())
    font = pygame.font.Font(FONT_PATH, 11)
    text = font.render(f"FPS: {fps}", True, (255, 255, 255), BG_COLOR)
    text_width = text.get_rect().width
    display.blit(text, (WINDOW_WIDTH - MARGIN - text_width, WINDOW_HEIGHT - MARGIN))



from .particle import Particle
from . import particle_data
from random import *


class Fixed(Particle):
    def __init__(
            self,
            col, row,
            vel_x, vel_y,
            acc_x, acc_y,
            temp, temp_freeze, temp_boil,
            density,
            color,
            name,
            flammability,
            state):
        super().__init__(
            col, row,
            vel_x, vel_y,
            acc_x, acc_y,
            temp, temp_freeze, temp_boil,
            density,
            color,
            name,
            flammability,
            state)

    def clone(self, col, row):
        return Fixed(
            col, row,
            self._vel_x, self._vel_y,
            self._acc_x, self._acc_y,
            self._temp, self._temp_freeze, self._temp_boil,
            self._density,
            self._color,
            self._name,
            self._flammability,
            self._state)

    def update_on_tick(self, driver, grid):
        if self.name == "Water Generator":
            if not grid.exists((self.col + 1, self.row)):
                driver.add(particle_data.template_water.clone(self.col + 1, self.row))

        if self.name == "bush":
            growth = random()
            if growth < 0.003 and self.row > 6:
                if not grid.exists((self.col, self.row - 1)):
                    driver.add(particle_data.template_bush.clone(self.col, self.row - 1))
            growth = random()
            if growth < 0.001:
                if not grid.exists((self.col + 1, self.row)):
                    driver.add(particle_data.template_bush.clone(self.col + 1, self.row))
            growth = random()
            if growth < 0.001:
                if not grid.exists((self.col - 1, self.row)):
                    driver.add(particle_data.template_bush.clone(self.col - 1, self.row))

        if self.name == "vine":
            growth = random()
            if growth < 0.03 and self.row > 6:
                if not grid.exists((self.col, self.row - 1)):
                    driver.add(particle_data.template_vine.clone(self.col, self.row - 1))
            growth = random()
            if growth < 0.001 and self.row % 7 == 0:
                if not grid.exists((self.col + 1, self.row)):
                    driver.add(particle_data.template_vine.clone(self.col + 1, self.row))
            growth = random()
            if growth < 0.001 and self.row % 7 == 0:
                if not grid.exists((self.col - 1, self.row)):
                    driver.add(particle_data.template_vine.clone(self.col - 1, self.row))

        if self._needs_update is False:
            return

        near_list = grid.get_near((self._col, self._row))

        # Heat transfer
        for particle in near_list:
            temp_diff = (self._temp - particle.temp) / 50
            particle.update_temp(particle.temp + temp_diff)
            self.update_temp(self._temp - temp_diff)

        # All fixed solids above boil temp -> gas
        if self._temp_boil <= self._temp:
            self.boil(driver, grid, particle_data.template_steam.clone(self._col, self._row))

        # Ice melts into water
        if self.name == "ice" and self._temp_freeze <= self._temp:
            oldtemp = self._temp
            self.melt(driver, grid, particle_data.template_water.clone(self._col, self._row))
            self.update_temp(oldtemp)

        # Basalt and metal melts into lava
        if (self.name == "basalt" or self.name == "metal") and self._temp_freeze <= self._temp:
            oldtemp = self._temp
            self.melt(driver, grid, particle_data.template_lava.clone(self._col, self._row))
            self.update_temp(oldtemp)

        # Wood burns
        if self.name == "wood" and self._temp_freeze <= self._temp:
            oldtemp = self._temp
            self.melt(driver, grid, particle_data.template_fire.clone(self._col, self._row))
            self.update_temp(oldtemp)

        # Molten metal -> lava
        if self.name == "metal" and self._temp_freeze <= self._temp:
            oldtemp = self._temp
            self.melt(driver, grid, particle_data.template_lava.clone(self._col, self._row))
            self.update_temp(oldtemp)

        self._needs_update = False



import pygame
from .config import FONT_PATH, BG_COLOR
from .particle_data import ELEMENTS


class ElementMenu:
    pygame.font.init()
    FONT = pygame.font.Font(FONT_PATH, 11)
    FONT_COLOR = (255, 255, 255)
    BUTTON_SIZE = 18
    MARGIN = 6

    def __init__(self, surface, x, y, width, mode):
        self._surface = surface
        self._x = x
        self._y = y
        self._width = width
        self._height = 150
        self._mode = mode
        self.element_buttons = []
        self.draw(True)

    # Creates the menu by building each section based on particle_data
    # Create flag determines whether button should be instantiated (false for redrawing section labels)
    def draw(self, create):
        sect_x = self._x
        sect_y = self._y
        sect_w = self._width / 2 - self.MARGIN
        for category in ELEMENTS.keys():
            self.create_section(self._surface, sect_x, sect_y, sect_w, category, ELEMENTS[category], create)
            sect_x += sect_w + self.MARGIN
            if sect_x >= self._x + self._width:
                sect_x = self._x
                sect_y += 50

    # First checks if a button was clicked, then changes corresponding button to active
    def update(self, driver, x, y):
        for b in self.element_buttons:
            if b.contains(x, y):
                for button in self.element_buttons:
                    button.set_active(False)
                    if button.contains(x, y) and button.unlocked:
                        button.set_active(True)
                        driver.set_current_element(button.get_element())
                    button.update()

    def draw_tooltip(self, mouse_x, mouse_y):
        elem_menu_rect = pygame.Rect(self._x, self._y, self._width, self._height)
        pygame.draw.rect(self._surface, BG_COLOR, elem_menu_rect)
        self.draw(False)

        hovered = None
        for button in self.element_buttons:
            button.update()
            if button.contains(mouse_x, mouse_y) and button.unlocked:
                hovered = button
        if hovered is None:
            return

        font = pygame.font.Font(FONT_PATH, 11)
        label = font.render(f"{hovered.particle.name.upper()}", True, (255, 255, 255), (0, 0, 0))
        x_offset, y_offset = 15, 10
        self._surface.blit(label, (mouse_x + x_offset, mouse_y - y_offset))

    def contains(self, x, y):
        if x < self._x or self._x + self._width < x:
            return False
        if y < self._y or self._y + self._height < y:
            return False
        return True

    def create_section(self, surface, x, y, width, title, elements, create):
        t = self.FONT.render(title, True, self.FONT_COLOR)
        surface.blit(t, (x, y))

        left = x
        top = y + t.get_height() + self.MARGIN/2

        for e in elements:
            if create:
                self.element_buttons.append(self.ElementButton(surface, left, top, self.BUTTON_SIZE,
                                                               self.BUTTON_SIZE, e))
            left += self.BUTTON_SIZE + self.MARGIN
            if left + self.BUTTON_SIZE > x + width:
                left = x
                top += self.BUTTON_SIZE + self.MARGIN

        return top + self.BUTTON_SIZE - self._y

    def discovery_init(self, undiscovered):
        for button in self.element_buttons:
            if button.get_element().name in undiscovered:
                button._unlocked = False
            else:
                button._unlocked = True
            button.update()

    class ElementButton:
        ACTIVE_COLOR = (255, 0, 0)

        def __init__(self, surface, x, y, width, height, template_particle):
            self._surface = surface
            self._x = x
            self._y = y
            self._width = width
            self._height = height
            self._particle = template_particle
            self._active = False
            self._enabled = True
            self._unlocked = True
            self.update()

        # Redraws button based on particle and if unlocked/enabled/active
        def update(self):
            button = pygame.Rect(self._x, self._y, self._width, self._height)

            # If element is locked draw a black square with a question mark, otherwise draw square with element color
            if not self._unlocked:
                pygame.draw.rect(self._surface, (0, 0, 0), button)
                font = pygame.font.Font(FONT_PATH, 11)
                q_mark = font.render("?", False, (255, 255, 255))
                self._surface.blit(q_mark, (self._x + 4, self._y + 1))
            else:
                pygame.draw.rect(self._surface, self._particle.color, button)

            # If element is not enabled, draw a semi-transparent white square over it
            if not self._enabled:
                s = pygame.Surface((self._width, self._height), pygame.SRCALPHA)
                s.fill((150, 150, 150, 180))
                self._surface.blit(s, (self._x, self._y))

            # If element is currently active/selected, draw a red line around button
            if self._active:
                pygame.draw.lines(self._surface, self.ACTIVE_COLOR, True,
                                  ((self._x, self._y),
                                   (self._x + self._width-1, self._y),
                                   (self._x + self._width-1, self._y + self._height-1),
                                   (self._x, self._y + self._height-1)))

            return button

        def get_element(self):
            return self._particle

        # def set_active(self):
        #     self._active = True
        #
        # def set_inactive(self):
        #     self._active = False

        def get_enabled(self):
            return self._enabled

        def set_enabled(self, status):
            self._enabled = status

        enabled = property(get_enabled, set_enabled)

        def get_lock(self):
            return self._unlocked

        def set_lock(self, status):
            self._unlocked = status

        unlocked = property(get_lock, set_lock)

        def get_active(self):
            return self._active

        def set_active(self, active):
            self._active = active

        active = property(get_active, set_active)

        def get_particle(self):
            return self._particle

        def set_particle(self, particle):
            self._particle = particle

        particle = property(get_particle, set_particle)

        def contains(self, x, y):
            if x < self._x or self._x + self._width < x:
                return False

            if y < self._y or self._y + self._height < y:
                return False

            return True



import abc
import math

from .grid_object import GridObject
from .solid_body_particle import SolidBodyParticle


class SolidBody(GridObject):
    def __init__(
            self,
            id_color_dict,
            id_2d_list,
            width, height,
            col, row,
            vel_x, vel_y,
            acc_x, acc_y,
            temp, temp_freeze, temp_boil,
            density,
            color,
            name,
            flammability,
            state):

        super().__init__(
            col, row,
            vel_x, vel_y,
            acc_x, acc_y,
            temp, temp_freeze, temp_boil,
            density,
            color,
            name,
            flammability,
            state)

        self._id_color_dict = id_color_dict
        self._id_2d_list = id_2d_list
        self._width = width
        self._height = height
        self._particles = [[None for _ in range(width)] for _ in range(height)]

        c_width = int(math.floor(width * 0.5))
        c_height = int(math.floor(height * 0.5))

        self._col -= c_width
        self._row -= c_height

        for y in range(len(id_2d_list)):
            for x in range(len(id_2d_list[y])):
                curr_color = id_color_dict[id_2d_list[y][x]]
                if curr_color is None:
                    self._particles[y][x] = None
                else:
                    self._particles[y][x] = SolidBodyParticle(
                        self._col + x, self._row + y,
                        self._vel_x, self._vel_y,
                        self._acc_x, self._acc_y,
                        self._temp, self._temp_freeze, self._temp_boil,
                        self._density,
                        curr_color,
                        self._name,
                        self._flammability,
                        self._state)

    @abc.abstractmethod
    def clone(self, col, row):
        pass

    def emplace(self, driver):
        can_add = True

        for row in range(self._height):
            for col in range(self._width):
                if self._particles[row][col] is not None:
                    pos = (self._col + col, self._row + row)
                    if not driver.grid.is_in_bounds(pos) or driver.grid.exists(pos):
                        can_add = False

        if can_add:
            driver.add_solid_body(self)
            for row in range(self._height):
                for col in range(self._width):
                    if self._particles[row][col] is not None:
                        driver.add(self._particles[row][col])

    @abc.abstractmethod
    def update_on_tick(self, driver, grid):
        pass

    def contains(self, particle):
        for row in range(self._height):
            if particle in self._particles[row]:
                return True
        return False

    def empty(self):
        return len(self._particles) == 0



import math
import os
import pygame

SANDBOX_WIDTH = 800
SANDBOX_HEIGHT = 600

MARGIN = 12

SANDBOX_X = MARGIN
SANDBOX_Y = MARGIN * 2

WINDOW_WIDTH = SANDBOX_WIDTH + (MARGIN * 2)
WINDOW_HEIGHT = SANDBOX_HEIGHT + (MARGIN * 4) + math.floor(SANDBOX_HEIGHT * 0.3)

MENU_WIDTH = 828
MENU_HEIGHT = 828

MODE = "A"

FPS = 60
PARTICLE_SIZE = 4

FONT_PATH = os.path.join(os.path.dirname(__file__), 'fonts', 'RetroGaming.ttf')  # "fonts/RetroGaming.ttf"
FONT_COLOR = (255, 255, 255)

BG_COLOR = (33, 33, 33)
SANDBOX_COLOR = (0, 0, 0)

TOMENU_EVENT_TYPE = pygame.USEREVENT + 1
TOMENU_EVENT = pygame.event.Event(TOMENU_EVENT_TYPE,)



import pickle
import tkinter as tk
from tkinter import filedialog
from .config import *
from .grid import Grid, px_to_cell
from .painter import Painter
from .particle_data import *
from .particle import Particle
from .gas import Gas


def print_state_message(screen, text):
    # Clear previous message
    clear_surf = pygame.Surface((SANDBOX_WIDTH, (WINDOW_HEIGHT - SANDBOX_HEIGHT) / 2))
    clear_surf.fill(BG_COLOR)
    screen.blit(clear_surf, clear_surf.get_rect().move(
        (SANDBOX_WIDTH / 2 - clear_surf.get_width() / 2, WINDOW_HEIGHT - 1.25 * SANDBOX_Y)))

    # Print new message
    print_font = pygame.font.Font(FONT_PATH, 20)
    text_rect = print_font.render(text, False, (255, 255, 255))
    screen.blit(text_rect, text_rect.get_rect().move(
        (SANDBOX_WIDTH / 2 - text_rect.get_width() / 2, WINDOW_HEIGHT - 1.25 * SANDBOX_Y)))


class Driver:
    def __init__(self, mode, element_menu, display):
        self.__particles = []
        self.__solid_bodies = []
        self.__grid = Grid()
        self.__painter = Painter(template_sand)
        self._mode = mode
        self._tool = "ADD"
        self._size = 1
        self._tool_use = False
        self._shape_start = (0, 0)
        self._shape_end = (0, 0)
        self._shape_active = False
        self.undiscovered = []
        self.__element_menu = element_menu
        self.__display = display

        # Initializes and clears Tkinter window, allows for filedialog
        root = tk.Tk()
        root.withdraw()

        if mode == 'LOAD':
            self.load_state()

    """
        add adds the specified particle both the particle list and the Grid the particle into the grid.
        This is the method that should be called when adding a particle to the Grid, NOT Grid.emplace
    """
    def add(self, particle):
        self.__particles.append(particle)
        self.__grid.emplace(particle)

    def add_list(self, particle_list):
        for particle in particle_list:
            if particle is not None:
                self.add(particle)

    def add_solid_body(self, solid_body):
        self.__solid_bodies.append(solid_body)

    def delete(self, particle):
        try:
            self.__particles.remove(particle)
            self.__grid.remove(particle)
            for p in self.__grid.get_near((particle.col, particle.row)):
                p.force_update()
        except ValueError:
            pass

    def get_tool(self):
        return self._tool

    def set_tool(self, tool):
        self._tool = tool

    def get_size(self):
        return self._size

    # Boolean for if tool is active
    def set_tool_use(self, status):
        self._tool_use = status

    def end_active_shape(self):
        self._shape_active = False

    # Changes brush size with min, max value
    def set_size(self, value):
        self._size += value
        if self._size < 1:
            self._size = 1
        if self._size > 6:
            self._size = 6

    def clear_sandbox(self):
        self.__particles.clear()
        self.__grid = Grid()

    def clear_element(self):
        selected_elem = None
        for elem in self.__element_menu.element_buttons:
            if elem.active:
                selected_elem = elem.particle
                break
        if selected_elem is None:
            return
        while selected_elem.name in [elem.name for elem in self.__particles]:
            for particle in self.__particles:
                if particle.name == selected_elem.name:
                    try:
                        self.__particles.remove(particle)
                        self.__grid.remove(particle)
                        for p in self.__grid.get_near((particle.col, particle.row)):
                            p.force_update()
                    except ValueError:
                        pass

    # Draws gray square outline instead of mouse, clips so not drawn outside sandbox
    def draw_tool_outline(self, pos, sandbox, display):
        line_color = (100, 100, 100)

        if self._tool == "LINE" and self._shape_active:
            r = math.atan2((pos[1] - self._shape_start[1]), (pos[0] - self._shape_start[0])) + math.radians(90)
            width = self._size * PARTICLE_SIZE

            s1 = sandbox.clipline(self._shape_start[0], self._shape_start[1], pos[0], pos[1])
            s2 = sandbox.clipline(pos[0], pos[1], pos[0] + (width * math.cos(r)), pos[1] + (width * math.sin(r)))
            s3 = sandbox.clipline(self._shape_start[0], self._shape_start[1], self._shape_start[0] +
                                  (width * math.cos(r)), self._shape_start[1] + (width * math.sin(r)))
            s4 = sandbox.clipline(self._shape_start[0] + (width * math.cos(r)), self._shape_start[1] +
                                  (width * math.sin(r)), pos[0] + (width * math.cos(r)), pos[1] +
                                  (width * math.sin(r)))

        elif self._tool == "RECT" and self._shape_active:
            (x1, y1) = (self._shape_start[0], self._shape_start[1])
            (x2, y2) = (pos[0], pos[1])

            s1 = sandbox.clipline(x1, y1, x1, y2)
            s2 = sandbox.clipline(x1, y2, x2, y2)
            s3 = sandbox.clipline(x2, y2, x2, y1)
            s4 = sandbox.clipline(x2, y1, x1, y1)

        elif self._tool == "OVAL" and self._shape_active:
            (x1, y1) = (self._shape_start[0], self._shape_start[1])
            (x2, y2) = (pos[0], pos[1])
            width, height = abs(x2 - x1), abs(y2 - y1)

            shape = pygame.Rect(min(x1, x2), min(y1, y2), width, height)
            pygame.draw.ellipse(display, line_color, shape, 1)

            return
        else:
            if self._tool == "INSPECT":
                size = PARTICLE_SIZE
                x, y = px_to_cell(pos[0]), px_to_cell(pos[1])

                font = pygame.font.Font(FONT_PATH, 11)
                if self.__grid.exists((x, y)):
                    particle = self.__grid.get((x, y))
                    label = font.render(f"{particle.name}: {x}, {y} Temp: {str(round(particle.temp, 1))} C",
                                        True, (255, 255, 255), (0, 0, 0))
                else:
                    label = font.render(f"Empty: {x}, {y}", True, (255, 255, 255), (0, 0, 0))

                # Adjust label location depending on proximity to edge of Sandbox
                x_offset, y_offset = 10, 5
                if pos[0] > SANDBOX_WIDTH * 0.75:
                    x_offset = -label.get_width() - 10
                if pos[1] > SANDBOX_HEIGHT * 0.95:
                    y_offset = -label.get_height() - 5

                display.blit(label, (pos[0] + x_offset, pos[1] + y_offset))

            # Draw tool for add/delete
            else:
                size = self._size * PARTICLE_SIZE

            # Create lines for square, cropped by Sandbox
            s1 = sandbox.clipline(pos[0], pos[1], pos[0] + size, pos[1])
            s2 = sandbox.clipline(pos[0] + size, pos[1], pos[0] + size, pos[1] + size)
            s3 = sandbox.clipline(pos[0] + size, pos[1] + size, pos[0], pos[1] + size)
            s4 = sandbox.clipline(pos[0], pos[1] + size, pos[0], pos[1])

        # Draw tool outlines (if they exist)
        if s1:
            pygame.draw.line(display, line_color, s1[0], s1[1])
        if s2:
            pygame.draw.line(display, line_color, s2[0], s2[1])
        if s3:
            pygame.draw.line(display, line_color, s3[0], s3[1])
        if s4:
            pygame.draw.line(display, line_color, s4[0], s4[1])

    def start_shape(self, pos):
        self._shape_start = pos
        self._shape_active = True

    def end_line(self, pos):
        self._shape_end = pos
        self._shape_active = False
        self.draw_line()

    def end_rect(self, pos):
        self._shape_end = pos
        self._shape_active = False
        self.draw_rect()

    def end_oval(self, pos):
        self._shape_end = pos
        self._shape_active = False
        self.draw_oval()

    # Ellipse collision detection solution from:
    # https://stackoverflow.com/questions/59971407/how-can-i-test-if-a-point-is-in-an-ellipse
    def draw_oval(self):
        width, height = abs(self._shape_end[0] - self._shape_start[0]), abs(self._shape_end[1] - self._shape_start[1])
        shape = pygame.Rect(min(self._shape_start[0], self._shape_end[0]),
                            min(self._shape_start[1], self._shape_end[1]), width, height)

        x1, y1, x2, y2 = shape.left, shape.top, shape.right, shape.bottom
        cx, cy = shape.centerx, shape.centery
        a = width // 2
        b = height // 2

        if a == 0 or b == 0:
            return
        scale = a / b

        for i in range(x1, x2 + 1):
            for j in range(y1, y2 + 1):
                dx = i - cx
                dy = (j - cy) * scale
                if dx*dx + dy*dy <= a*a:
                    (px, py) = (px_to_cell(i), px_to_cell(j))
                    if self.__grid.is_in_bounds([px, py]) and self.__grid.exists([px, py]) is False:
                        self.add(self.__painter.get_template_particle().clone(px, py))

    def draw_line(self):
        (p1, p2) = (px_to_cell(self._shape_end[0]), px_to_cell(self._shape_end[1]))

        if not self.__grid.is_in_bounds([p1, p2]):
            return

        r = math.atan2((self._shape_end[1] - self._shape_start[1]), (self._shape_end[0] - self._shape_start[0]))
        r2 = r + math.radians(90)
        d = math.floor(math.sqrt((self._shape_end[0] - self._shape_start[0])**2 +
                                 (self._shape_end[1] - self._shape_start[1])**2))
        w = self._size * PARTICLE_SIZE

        for i in range(d):
            for j in range(w):
                x = self._shape_start[0] + (i * math.cos(r)) + (j * math.cos(r2))
                y = self._shape_start[1] + (i * math.sin(r)) + (j * math.sin(r2))

                (px, py) = (px_to_cell(x), px_to_cell(y))
                if self.__grid.exists([px, py]) is False:
                    self.add(self.__painter.get_template_particle().clone(px, py))

    def draw_rect(self):
        (p1x, p1y) = (px_to_cell(self._shape_start[0]), px_to_cell(self._shape_start[1]))
        (p2x, p2y) = (px_to_cell(self._shape_end[0]), px_to_cell(self._shape_end[1]))

        if p2x < p1x:
            temp = p1x
            p1x = p2x
            p2x = temp

        if p2y < p1y:
            temp = p1y
            p1y = p2y
            p2y = temp

        if not self.__grid.is_in_bounds([p2x, p2y]):
            return

        for i in range(p1x, p2x+1):
            for j in range(p1y, p2y+1):
                if self.__grid.exists([i, j]) is False:
                    self.add(self.__painter.get_template_particle().clone(i, j))

    # For each particle, update its position. Then, apply tool if active
    def update_particles(self, mouse):
        for solid_body in self.__solid_bodies:
            if solid_body.empty():
                self.__solid_bodies.remove(solid_body)
            else:
                solid_body.update_on_tick(self, self.__grid)

        for particle in self.__particles:
            particle.update_on_tick(self, self.__grid)

            if particle.is_live is False or self.__grid.exists([particle.col, particle.row]) is False:
                try:
                    self.__particles.remove(particle)
                except ValueError:
                    pass

            if particle.name == "Water Generator":
                particle.force_update()

        if self._tool_use:
            if self._tool == "ADD" or self._tool == "DELETE" or self._tool == "ERASE":
                self.__painter.use_tool(mouse, self, self.__grid)

    def get_current_element(self):
        return self.__painter.get_template_particle()

    def set_current_element(self, new):
        self.__painter.set_template_particle(new)

    # Draws particles in sandbox, also checks for new elements in Discovery Mode
    def render(self, screen):
        for particle in self.__particles:
            particle.render(screen)

            if self._mode == "DISCOVERY" and len(self.undiscovered) > 0 and particle.name in self.undiscovered:
                for e in self.__element_menu.element_buttons:
                    if e.get_element().name == particle.name:
                        e._unlocked = True
                        e.update()
                        for elem in self.undiscovered:
                            if elem == particle.name:
                                self.undiscovered.remove(elem)
                                break

                        click = False
                        while not click:
                            font = pygame.font.Font(FONT_PATH, 30)
                            alert = font.render("%s DISCOVERED!" % particle.name.upper(), False, (255, 255, 255))
                            alert_rect = alert.get_rect()
                            alert_rect.center = (SANDBOX_WIDTH / 2, SANDBOX_HEIGHT / 2)
                            screen.blit(alert, alert_rect)

                            font2 = pygame.font.Font(FONT_PATH, 18)
                            alert2 = font2.render("(Click to Continue)", False, (255, 255, 255))
                            alert2_rect = alert2.get_rect()
                            alert2_rect.center = (SANDBOX_WIDTH / 2, (SANDBOX_HEIGHT / 2) + 50)
                            screen.blit(alert2, alert2_rect)

                            pygame.display.update()

                            for event in pygame.event.get():
                                if event.type == pygame.MOUSEBUTTONDOWN:
                                    click = True
                                if event.type == pygame.KEYDOWN:
                                    click = True
                        return

    def save_state(self):
        for particle in self.__particles:
            if isinstance(particle, Gas):
                particle.save_lifespan()
        data = {
            'particles': self.__particles,
            'undiscovered': self.undiscovered,
            'mode': self._mode
        }
        file_path = filedialog.asksaveasfilename(defaultextension='.pickle',
                                                 filetypes=[('Pickle Files', '*.pickle')])
        if file_path == '':  # if cancel button clicked
            return
        with open(file_path, 'wb') as file:
            pickle.dump(data, file)
        print_state_message(self.__display, 'Saved!')

    def load_state(self):
        file_path = filedialog.askopenfilename(title='Select a File',
                                               filetypes=[('Pickle Files', '*.pickle')])
        if file_path == '':  # if cancel button clicked
            return
        with open(file_path, 'rb') as file:
            try:
                data = pickle.load(file)
                self.clear_sandbox()

                for particle in data['particles']:
                    if isinstance(particle, Particle):
                        if isinstance(particle, Gas):
                            particle.load_lifespan()
                        self.add(particle)
                    else:
                        raise ValueError

                if data['mode'] not in ["DISCOVERY", "SANDBOX"]:
                    raise ValueError
                elif data['mode'] == "DISCOVERY":
                    self.undiscovered = data['undiscovered']
                else:
                    self.undiscovered = []

                self.__element_menu.discovery_init(self.undiscovered)

                self._mode = data['mode']

                print_state_message(self.__display, 'Success: Loaded!')
            except:
                print_state_message(self.__display, 'Error: %s has invalid data!' % os.path.basename(file_path))

    @property
    def grid(self):
        return self.__grid



from .basic_solid_body import BasicSolidBody


class UFOSolidBody(BasicSolidBody):
    def __init__(
            self,
            id_color_dict,
            id_2d_list,
            width, height,
            col, row,
            vel_x, vel_y,
            acc_x, acc_y,
            temp, temp_freeze, temp_boil,
            density,
            color,
            name,
            flammability,
            state):

        super().__init__(
            id_color_dict,
            id_2d_list,
            width, height,
            col, row,
            vel_x, vel_y,
            acc_x, acc_y,
            temp, temp_freeze, temp_boil,
            density,
            color,
            name,
            flammability,
            state)

    def clone(self, col, row):
        return UFOSolidBody(
            self._id_color_dict,
            self._id_2d_list,
            self._width, self._height,
            col, row,
            self._vel_x, self._vel_y,
            self._acc_x, self._acc_y,
            self._temp, self._temp_freeze, self._temp_boil,
            self._density,
            self._color,
            self._name,
            self._flammability,
            self._state)

    def update_on_tick(self, driver, grid):
        can_move = self._update_position(grid)
        if not can_move:
            self._vel_x = -self._vel_x
